// (C) J. Ulasiewicz 2012 ----------------------------------------------
// Program typu SLAVE do odbioru ramek MODBUS ASCII 
// Obslugiwana jest funkcja 4 - odczyt rejestrow 16 bitowych
// Partnerem jest program modbus1_wys.c
// Domyslny port /dev/ser1
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#define PORT_COM1 "/dev/ser1"
#define SIZE 64
#define LF 10
#define CR 13
#define SOF ':'              /* ':' jest znacznikiem pocz. ramki */
#define DataSize   40  /* dlugosc pola danych  */
#define FrameSize  2*DataSize + 10   /* dlugosc ramki        */

/* Protokol MODBUS  ASCII Mode
  Struktura ramki ASCII
   +---------+----------+-----------+---------+--------+--------+
   I  Begin  I  Address I  Function I  Data   I   LRC  I   End  I
   I  1      I    2     I    2      I  2*N    I   2    I   2    I
   +---------+----------+-----------+---------+--------+--------+
   I  ':'    I  adr     I   funct   I xxxxxxx I  lrc   I CR LF  I
   +---------+----------+-----------+---------+--------+--------+

    Kazdy znak adr, funct i data reprezentowany jako dwa znaki HEX.
    LRC liczone dla pol adr, funct i data.
*/

char FrameBuf[FrameSize];          /* bufor na ramke    */
char Data[DataSize];               /* bufor roboczy  0  */
// Przykladowa odpowiedz na funkcje 4, wartosc danych 1 
//                       : 01 04 02  00 01 F8 <CR> <NL>    
char odp4[] = {0x3A, 0x30, 0x31, 0x30, 0x34, 0x30, 0x32, 0x30, 0x30, 0x30, 0x31, 0x46, 0x38, 0x0D, 0x0A};

unsigned char DigToHex(unsigned char d)
/* Konwersja liczb 0..F do znakow '0'..'F' */
{
  // if((d>=0) && (d<10))   return('0' + d);
  if(d<10)    return('0' + d);
  if((d>=10) && (d<=15)) return('A' + d - 10);
}

void CharToHex(char c, char x[3])
/* Convert char c to hex reprezentation x */
{
  unsigned char xh,xl;
  xl = (unsigned char)c & 0x0f;
  xh = ((unsigned char)c & 0xf0) >> 4;
  x[0] = DigToHex(xh); x[1] = DigToHex(xl); x[2] = 0;
}

unsigned char HexToDig(char c)
/* Konwersja znakow '0'..'F' do liczb 0..F */
{
  if((c>='0') && (c<='9')) return(c - '0');
  if((c>='A') && (c<='F')) return(c - 'A' + 10);
}

char HexToChar(char x[3])
/* Convert hex string x to char */
{
  unsigned char xh,xl,c;
  xh = HexToDig(x[0]);
  xl = HexToDig(x[1]);
  c = ((xh << 4) + xl);
  return (char) c;
}


unsigned char lrcgen(char *buf, int size)
/* Wyznaczanie LCR dla bufora buf i size znakow     */
{
 unsigned char sum = 0;
 int i;
 for(i=0;i<size;i++) sum = sum + buf[i];
 return(~sum + 1);
}



void SendFrm(int dev, int Adr, int Command, int Size, char *Data)
/* Send Command to port dev, unit Adr, Size char of Data      */
/* Uzyta zmienna globalna FrameBuf                            */
{
 unsigned char lrc;
 int i,cnt;
 char tmp[3];


 /* Computing of LCR ----------------*/
 FrameBuf[0] = (char)(Adr & 0x00FF);
 FrameBuf[1] = (char)(Command & 0x00FF);
 if( Size > 0)  for(i=0;i<Size;i++) FrameBuf[i+2] = Data[i];
 lrc = lrcgen(FrameBuf,Size + 2);
 /* Budowa ramki -------------------------------------------------*/
 FrameBuf[0] = ':';
 /* Address -----------------------------------------*/
 CharToHex((char) Adr,tmp);
 FrameBuf[1] = tmp[0]; FrameBuf[2] = tmp[1];
 /* Command -----------------------------------------*/
 CharToHex((char)Command,tmp);
 FrameBuf[3] = tmp[0]; FrameBuf[4] = tmp[1];
 cnt = 5;
 if( Size > 0 ) { /* Data ---------------------------*/
    for(i=0;i<Size;i++) {
      CharToHex(Data[i],tmp);
      FrameBuf[cnt] = tmp[0]; FrameBuf[cnt+1] = tmp[1];
      cnt++; cnt++;
    }
  }
 /* LRC ---------------------------------------*/
 CharToHex((char) lrc,tmp);
 FrameBuf[cnt] = tmp[0]; FrameBuf[cnt+1] = tmp[1]; cnt= cnt+2;
 /* End of Frame Marks ------------------------*/
 FrameBuf[cnt] = CR; cnt++;
 FrameBuf[cnt] = LF; cnt++;
 FrameBuf[cnt+1] = '\0';      /* <-- usunac   */
 /* Ramka gotowa --> wysylac  */
 printf("\nWyslano:  ");
 for(i=0;i<cnt;i++) {
    write(dev,&FrameBuf[i],1);
    printf("%c",FrameBuf[i]);
 }   
  printf("wyslano: %d znakow\n",cnt);
}


int main(int argc, char * argv[]) {
  int fd,speed, timeout,rd,res,i;
  int Adr,Cmd,Size;
  short int val = 0;
  struct termios term;
  char c, buf[SIZE];
  fd = open(PORT_COM1,O_RDWR);
  /* ustawienia predkosci transmitowania */
  speed=9600;
  timeout = 20; // 2 sek
  tcgetattr(fd,&term);
  term.c_cc[VFWD] = LF;
  tcsetattr( fd, TCSANOW, &term);
  tcflush(fd,TCIFLUSH);
  tcflush(fd,TCOFLUSH);
  do {
/*
      res = RecFrm(fd,&Adr,&Cmd,&Size,Data);
      if(res == 0)  {
          printf("Timeout\n");
         continue;
      }   Data[Size] = 0;
*/
     do {
        rd =  read(fd,&c,1);
        printf("%c",c);
        if(c == LF) break;
     } while(1);     

      Adr = 1; Cmd = 4; Size = 4;      
      printf("Odebrano adres: %d polecenie: %d rozmiar: %d %s \n",Adr,Cmd,Size,Data);
      switch(Cmd) {
      case 4: // Odczyt kanalu DA ----------
                  val++;
                  Data[0] = 0x02; // Liczba bajtow
                  Data[1] = val >>8;
                  Data[2] = val & 0x00FF;
                  printf("Dane: %02X %02X \n",Data[1],Data[2]);
                  Size = 3;
                  break;
     default: // Nie zaimplementowane --
                 printf("Funkcja nie zaimplementowana\n");
     }
     write(fd,odp4,sizeof(odp4));
    // SendFrm(fd,Adr,Cmd,Size,Data);                  
  } while(1);
  close(fd);
  return 0;
}
 
